/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.task;

import com.vividsolutions.jts.geom.Geometry;
import java.awt.geom.Point2D;
import java.util.List;
import nl.cloudfarming.client.field.model.ActivityField;
import nl.cloudfarming.client.field.model.TreatmentZone;
import nl.cloudfarming.client.lib.geotools.GeometryTools;
import nl.cloudfarming.client.model.EnterpriseIdProvider;
import nl.cloudfarming.client.model.ItemIdType;
import nl.cloudfarming.client.task.grid.GridGenerator;
import nl.cloudfarming.client.task.grid.TreatmentZoneGrid;
import nl.cloudfarming.client.task.util.GeometryUtil;

/**
 *
 * @author Maciek Dulkiewicz <maciek.dulkiewicz@ordina.nl>
 */
public abstract class ExportableTask extends Task {

    private ExportableTaskCapability exportableTaskCapability = new ExportableTaskCapability(this);
    public static final int TREATMENT_ZONE_LIMIT = 254;
    private TreatmentZoneGrid exportableGrid;
    public static final String PROP_GRID = "grid";
    public static final String PROP_WORK_WIDTH = "workWidth";
    public static final String MIME_TYPE = "application/x-agrosense-exportable-task";
    private boolean regenerateGrid = true;

    public ExportableTask(EnterpriseIdProvider eip, ItemIdType itemIdType) {
        this(eip, new Class[0], itemIdType);
    }

    public ExportableTask(EnterpriseIdProvider eip, Class[] requires, ItemIdType itemIdType) {
        super(eip, requires, itemIdType);
    }

    public abstract double getWorkWidth();

    @Override
    protected void setStatus(Status status) {
        super.setStatus(status);

        if (Status.COMPLETE.equals(status) && getLookup().lookup(ExportableTaskCapability.class) == null) {
            ic.add(exportableTaskCapability);
        } else if (Status.INCOMPLETE.equals(status) && getLookup().lookup(ExportableTaskCapability.class) != null) {
            ic.remove(exportableTaskCapability);
        }

    }

    public TreatmentZoneGrid getExportableGrid() {
        if (regenerateGrid()) {
            createExportableGrid();
            setRegenerateGrid(false);
        }
        return exportableGrid;
    }

    public boolean regenerateGrid() {
        return regenerateGrid;
    }

    public void setRegenerateGrid(boolean regenerate) {
        regenerateGrid = regenerate;
    }

    /**
     * This method creates and returns a {@link TreatmentZoneGrid} using the
     * treatment zones contained by the {@link TaskContext} of this task. The
     * number of treatment zones used in the TreatmentZoneGrid is limited by the
     * maximum defined at the top of this class.
     *
     * @return A TreatmentZoneGrid containing the maximum number of treatment
     * zones supported by the current ISOBUS implementation
     */
    public void createExportableGrid() {
        ActivityField activityField = getLookup().lookup(ActivityField.class);
        List<TreatmentZone> treatmentZones = activityField.getTreatmentZones();
        //FIXME: ISOBUS supports a maximum number of 255  treatment zones 
        // (including the initial empty one). In order to support more 
        // treatment zones they should probably be merged beforehand.
        List<TreatmentZone> maximumNumberOfTreatmentZones = getMaximumNumberOfTreatmentZones(treatmentZones);
        double workWidth = getWorkWidth();

        final double LATITUDE_WORKWIDTH_BEARING = 0;
        final double LONGITUDE_WORKWIDTH_BEARING = 90;
        // Create envelope of treatmentzones for this task.
        Geometry envelope = GeometryUtil.getEnvelope(maximumNumberOfTreatmentZones);

        // Calculate nr. of gridCells in width and height
        double envelopeMinX = GeometryUtil.getMinX(envelope);
        double envelopeMaxX = GeometryUtil.getMaxX(envelope);
        double envelopeWidth = envelopeMaxX - envelopeMinX;
        double envelopeMinY = GeometryUtil.getMinY(envelope);
        double envelopeMaxY = GeometryUtil.getMaxY(envelope);
        double envelopeHeight = envelopeMaxY - envelopeMinY;

        // Calculate nr of columns and rows
        Point2D latitudeWorkWidth = GeometryTools.getDestinationPoint(envelopeMinY, envelopeMinX, LATITUDE_WORKWIDTH_BEARING, workWidth);
        Point2D longitudeWorkWidth = GeometryTools.getDestinationPoint(envelopeMinY, envelopeMinX, LONGITUDE_WORKWIDTH_BEARING, workWidth);
        //switched x and y coordinate is the other way round
        double workWidthDegrees = longitudeWorkWidth.getX() - envelopeMinX;
        double workheightDegrees = latitudeWorkWidth.getY() - envelopeMinY;

        double columns = envelopeWidth / workWidthDegrees;
        double rows = envelopeHeight / workheightDegrees;

        int nrOfGridColumns = new Double(Math.ceil(columns)).intValue();
        int nrOfGridRows = new Double(Math.ceil(rows)).intValue();

        TreatmentZoneGrid grid = GridGenerator.generate(
                GeometryUtil.getCellWidth(envelope, nrOfGridColumns),
                GeometryUtil.getCellHeight(envelope, nrOfGridRows),
                GeometryUtil.getMinX(envelope),
                GeometryUtil.getMinY(envelope),
                nrOfGridRows,
                nrOfGridColumns,
                maximumNumberOfTreatmentZones);

        TreatmentZoneGrid oldGrid = exportableGrid;
        exportableGrid = grid;
        getPropertyChangeSupport().firePropertyChange(PROP_GRID, oldGrid, exportableGrid);
    }

    /**
     * This method extracts a subList of treatment zones from the passed list.
     * The subList has a maximum size equal to the defined maximum number of
     * treatment zones supported by the current ISOBUS implementation.
     *
     * @param treatmentZones The raw list of treatment zones the user wishes to
     * export
     * @return a subList of treatment zones no greater than the treatment zone
     * limit defined in this class
     */
    protected List<TreatmentZone> getMaximumNumberOfTreatmentZones(List<TreatmentZone> treatmentZones) {
        int subListEnd = TREATMENT_ZONE_LIMIT;
        if (treatmentZones.size() < TREATMENT_ZONE_LIMIT) {
            return treatmentZones;
        }
        return treatmentZones.subList(0, subListEnd);
    }
}
