/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.task;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.io.Serializable;
import nl.cloudfarming.client.model.*;
import org.openide.util.Exceptions;
import org.openide.util.Lookup;
import org.openide.util.Lookup.Result;
import org.openide.util.LookupEvent;
import org.openide.util.LookupListener;
import org.openide.util.lookup.AbstractLookup;
import org.openide.util.lookup.InstanceContent;
import org.openide.util.lookup.ProxyLookup;

/**
 * Parent class for all specific task implementations in AgroSense. It provides
 * rudimentary features for tasks and the underlying task context.
 *
 * @author Maciek Dulkiewicz <maciek.dulkiewicz@ordina.nl>
 */
public abstract class Task extends AbstractEntity implements Lookup.Provider, Serializable {

    static final String ACTIONS_PATH = "Task";
    public static final String DEFAULT_TASK_NAME = "My task";
    public static final String PROP_DESTROYED = "destroyed";
    public static final String PROP_NAME = "name";
    public static final String PROP_STATUS = "status";
    private Status status = Status.INCOMPLETE;
    private TaskContext taskContext = new TaskContext();
    private final Class[] requires;
    private transient PropertyChangeSupport propertyChange = new PropertyChangeSupport(this);
    private String name;
    public static Task NULLTASK = new NullTask();
    protected InstanceContent ic = new InstanceContent();
    private final AbstractLookup abstractLookup = new AbstractLookup(ic);
    private transient Lookup lookup = new ProxyLookup(abstractLookup, getContext());
    private final ContextListener contextListener = new ContextListener();

    /**
     * The constructor for the task class.
     *
     * @param requires An array of classes defining what the task needs in order
     * to get the 'complete' status.
     */
    public Task(EnterpriseIdProvider eip, Class[] requires, ItemIdType itemIdType) {
        super(eip, itemIdType);
        this.requires = requires;
        setName(DEFAULT_TASK_NAME + Lookup.getDefault().lookup(TaskService.class).getTaskCount());
        taskContext.lookupResult(Entity.class).addLookupListener(contextListener);
    }

    public Task(EnterpriseIdProvider eip, ItemIdType itemIdType) {
        this(eip, new Class[0], itemIdType);
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        String oldName = getName();
        this.name = name;
        getPropertyChangeSupport().firePropertyChange(PROP_NAME, oldName, this.name);
    }

    public Status getStatus() {
        return status;
    }

    /**
     * Determine the status of the task and return true when the task is
     * complete.
     *
     * @return true when task is complete, false if not
     */
    public boolean isComplete() {
        return Status.COMPLETE.equals(this.status);
    }

    public TaskContext getContext() {
        return taskContext;
    }

    @Override
    public Lookup getLookup() {
        return lookup;
    }

    private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
            in.defaultReadObject();
            lookup = new ProxyLookup(abstractLookup, getContext());
    }

    public Class[] getRequiredClasses() {
        return requires;
    }

    @Override
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        getPropertyChangeSupport().addPropertyChangeListener(listener);
    }

    @Override
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        getPropertyChangeSupport().removePropertyChangeListener(listener);
    }

    @Override
    protected PropertyChangeSupport getPropertyChangeSupport() {
        if (propertyChange == null) {
            propertyChange = new PropertyChangeSupport(this);
        }
        return propertyChange;
    }

    /**
     * Simple method firing a PropertyChange event indicating that the
     * TaskManager of this task should sever all connections to this task and
     * prepare it for garbage collection.
     */
    public void destroy() {
        getPropertyChangeSupport().firePropertyChange(PROP_DESTROYED, null, this);
    }

    protected void setStatus(Status status) {
        Status old = this.status;
        this.status = status;
        getPropertyChangeSupport().firePropertyChange(PROP_STATUS, old, status);
    }

    /**
     * Task status enumeration
     */
    public enum Status {

        /**
         * Task contains all required objects
         */
        COMPLETE,
        /**
         * Task does not yet contains all required objects
         */
        INCOMPLETE;
    }

    private final class ContextListener implements LookupListener, Serializable {

        @Override
        public void resultChanged(LookupEvent le) {
            Result result = (Result) le.getSource();
            int found = 0;
            for (int i = 0; i < Task.this.requires.length; i++) {
                for (Object o : result.allInstances()) {
                    if (o.getClass().equals(Task.this.requires[i])) {
                        found++;
                        break;
                    }
                }
            }
            if (found == Task.this.requires.length) {
                Task.this.setStatus(Status.COMPLETE);
            } else {
                Task.this.setStatus(Status.INCOMPLETE);
            }
        }
    }

    /**
     * This class is used as a placeholder for regular task objects. When there
     * is no reference to a real task in the TaskService, a NullTask is used
     * instead. This Class really can't do anything so whenever it is set and
     * actions are still performed on the TaskService nothing happens but no
     * exceptions are thrown either. Using this NullTask in the task module
     * prevents a lot of 'if(x == null)' statements.
     */
    private static class NullTask extends Task {

        public NullTask() {
            super(new AgroURI("OTHER/1/TSK/2"), ItemIdType.TSK);
        }

        @Override
        public void addPropertyChangeListener(PropertyChangeListener listener) {
        }

        @Override
        public void destroy() {
        }

        @Override
        public boolean equals(Object obj) {
            return false;
        }

        @Override
        public TaskContext getContext() {
            return new TaskContext();
        }

        @Override
        public Lookup getLookup() {
            return new TaskContext();
        }

        @Override
        public String getName() {
            return "NULLTASK";
        }

        @Override
        public Class[] getRequiredClasses() {
            return new Class[0];
        }

        @Override
        public Status getStatus() {
            return Status.INCOMPLETE;
        }

        @Override
        public boolean isComplete() {
            return false;
        }

        @Override
        public void removePropertyChangeListener(PropertyChangeListener listener) {
        }

        @Override
        public void setName(String name) {
        }

        @Override
        protected void setStatus(Status status) {
        }
    }
}
