/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.task;

import java.io.Serializable;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import nl.cloudfarming.client.model.AgroURI;
import nl.cloudfarming.client.model.Entity;
import nl.cloudfarming.client.model.ItemIdType;
import org.openide.util.Lookup;
import org.openide.util.LookupEvent;
import org.openide.util.LookupListener;
import org.openide.util.lookup.AbstractLookup;
import org.openide.util.lookup.InstanceContent;

/**
 * This class serves as a context for Task objects. A context is a collection of
 * elements of a task. This implementation uses the URIs of entities used in
 * AgroSense to maintain references to items placed in a context.
 *
 * @author Maciek Dulkiewicz <maciek.dulkiewicz@ordina.nl>
 */
public class TaskContext extends AbstractLookup {

    public static final String PROP_CONTEXT = "context";
    private volatile List<LookupListener> lookupListeners;
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.task");
    private final InstanceContent content;
    private volatile Map<AgroURI, Entity> entityInstances;
    //WARNING do not add items to uris, add them to entityInstances instead
    private HashSet<AgroURI> uris;

    public TaskContext(InstanceContent content) {
        super(content);
        this.content = content;
        entityInstances = new ContextMap(content);
        uris = new HashSet(entityInstances.keySet());
        lookupListeners = new ArrayList<>();
    }

    public TaskContext() {
        this(new InstanceContent());
    }

    /**
     * Method used to find all instances of a specific class in the context.
     * This method will use both the entityInstances and uris maps to find all
     * instances of a specific class that have been added to this context. It
     * iterates through all uris in the uris map to find all uris that have a
     * ItemIdType equal to the one on the class passed as parameter. Each uri
     * with the specific itemIdType is then fetched from the entityInstances map
     * which contains known instances of some classes. If the entityInstances
     * map does not contain a reference to an instance of a class we need, the
     * entityManager is called to find an instance. All found instances are
     * added to a list an returned.
     *
     * @param <T>
     * @param clazz The class you wish to find all instances of
     * @return A list containing all found instances of the class passed as
     * parameter
     */
    @Override
    public <T> Collection<? extends T> lookupAll(Class<T> clazz) {
        if (clazz == Entity.class) {
            return (Collection<? extends T>) entityInstances.values();
        }

        Collection entitiesForItemIdType = new ArrayList<>();
        ItemIdType itemIdType = ItemIdType.fromClass(clazz);

        for (AgroURI uri : uris) {
            if (uri.getItemIdType().equals(itemIdType)) {
                if (entityInstances.get(uri) != null) {
                    entitiesForItemIdType.add(entityInstances.get(uri));
                } else {
                    //TODO: we need an entityManager to find objects by URIs
                    Entity entity = null;
                    //Entity entity = EntityManager.findByURI(uri);
                    if (entity != null) {
                        entityInstances.put(entity.getURI(), entity);
                        entitiesForItemIdType.add(entity);
                    } else {
                        LOGGER.log(Level.WARNING, "Unable to find entity by URI: {0}", uri);
                    }
                }
            }
        }
        return entitiesForItemIdType;
    }

    @Override
    public <T> Result<T> lookupResult(final Class<T> clazz) {
        Result result = new Result() {

            @Override
            public void addLookupListener(LookupListener l) {
                lookupListeners.add(l);
            }

            @Override
            public void removeLookupListener(LookupListener l) {
                lookupListeners.remove(l);
            }

            @Override
            public Collection allInstances() {
                return lookupAll(clazz);
            }

            @Override
            public Set allClasses() {
                Set<Class> classes = new HashSet<>();
                for (Entity entity : entityInstances.values()) {
                    classes.add(entity.getClass());
                }
                return classes;
            }
        };
        return result;
    }

    private void notifyLookupListeners(Lookup.Result result, String propertyName) {
        for (LookupListener listener : lookupListeners) {
            listener.resultChanged(new LookupEvent(result));
        }
    }

    /**
     * This method returns all entities contained by this context. Getting all
     * entity objects by calling lookupAll is not possible because the lookupAll
     * method tries to find an itemIdType for the class passed as parameter. The
     * entity class does not have a valid itemIdType so it is never processed by
     * the lookupAll method.
     *
     * @return Collection containing all entity objects of this context
     */
    public Collection<Entity> getAllEntities() {
        return entityInstances.values();
    }

    /**
     * This method adds an entity to both maps of the context. The URI of the
     * item together with the parent the item belongs to are added to the uris
     * map so they can be serialised later. The project is required because
     * entityManagers need it to determine where to look for instances of a
     * class. The item is also added to the entityInstances map, together with
     * its URI
     *
     * @param item The item you wish to add to the context
     * @param project The parent project of the item you wish to add to the
     * context
     */
    public void addItem(Entity item) {
        if (item.getURI() != null) {
            entityInstances.put(item.getURI(), item);
            uris.add(item.getURI());
            notifyLookupListeners(lookupResult(Entity.class), PROP_CONTEXT);
            LOGGER.log(Level.FINEST, "Added entity to context of active task: {0}", item);
        } else {
            LOGGER.log(Level.FINEST, "Tried to add entity to context but had no valid URI: {0}", item);
        }
    }

    /**
     * This method is used to remove an item from the context. It removes the
     * URI of the item from the uris map and makes sure the reference to the
     * item (if there was one) is removed from the entityInstances map.
     *
     * @param item The item you wish to remove from the context
     */
    public void removeItem(Entity item) {
        uris.remove(item.getURI());
        entityInstances.remove(item.getURI());
        notifyLookupListeners(lookupResult(Entity.class), PROP_CONTEXT);
    }

    /**
     * Returns the total number of items contained by this context. This is
     * determined by counting the number of uris in the uris map.
     *
     * @return
     */
    public int getItemCount() {
        return uris.size();
    }

    private final class ContextMap extends HashMap implements Serializable {

        private final InstanceContent content;

        public ContextMap(InstanceContent content) {
            this.content = content;
        }

        @Override
        public Object put(Object key, Object value) {
            content.add(value);
            return super.put(key, value);
        }
    }
}
