/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.task;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import nl.cloudfarming.client.model.Entity;
import org.openide.util.Lookup;
import org.openide.util.lookup.ServiceProvider;

/**
 *
 * @author Maciek Dulkiewicz <maciek.dulkiewicz@ordina.nl>
 */
@ServiceProvider(service = TaskService.class)
public final class TaskService implements PropertyChangeListener {

    public static final String PROP_ADD = "added";
    public static final String PROP_REMOVE = "removed";
    public static final String PROP_ITEM_ADD = "itemAdded";
    public static final String PROP_ITEM_REMOVE = "itemRemoved";
    private List<Task> taskList = new ArrayList<>();
    private Task activeTask = null;
    private PropertyChangeSupport propertyChange = new PropertyChangeSupport(this);

    /**
     * Adds a task to the general TaskList but doesn't make the task active.
     *
     * @param task Task to be added to the TaskList
     */
    public void addTask(Task task) {
        taskList.add(task);
        propertyChange.firePropertyChange(PROP_ADD, null, task);
    }

    /**
     * Adds a task to the general TaskList and sets the task as the currently
     * active task.
     *
     * @param task Task to be added to the TaskList
     */
    public void addAndActivateTask(Task task) {
        addTask(task);
        activeTask = task;
    }

    /**
     * Removes the given task from the general TaskList and makes sure the
     * active task value in TaskList is correctly set if the given task was the
     * active task.
     *
     * @Param task Task to be removed from the TaskList
     */
    public void removeTask(Task task) {
        taskList.remove(task);
        propertyChange.firePropertyChange(PROP_REMOVE, task, null);
    }

    /**
     * Sets the given task as the active task in the TaskList.
     *
     * @param task Task to be set as the active task
     */
    public void activateTask(Task task) {
        activeTask = task;
    }

    /**
     * Unsets the given task as the active task in the TaskList.
     *
     * @param task Task to be unset as active task
     */
    public void deactivateTask(Task task) {
        activeTask = null;
    }

    /**
     * Returns the number of tasks currently in the TaskList. This method counts
     * ALL tasks, active and inactive.
     *
     * @return The number of tasks in the TaskList
     */
    public int getTaskCount() {
        return taskList.size();
    }

    /**
     * Returns the Task with a specific index in the TaskList.
     *
     * @param taskIndex The index of the Task that should be returned
     * @return The task object held at the given index of the TaskList or null
     */
    public Task getTask(int taskIndex) {
        if (taskIndex < taskList.size()) {
            return taskList.get(taskIndex);
        }
        return null;
    }

    /**
     * Returns the Task object currently set as active in the TaskList.
     *
     * @return The Task object marked as active in the TaskList or null if there
     * is no active task.
     */
    public Task getActiveTask() {
        return activeTask;
    }

    public List<Task> getTasks() {
        return taskList;
    }

//    /**
//     * Uses the lookup to find all subclasses of the TaskManager class, adds
//     * this TaskService class as a listener to each manager and makes sure each
//     * manager loads the tasks it is responsible for. Each TaskManager fires a
//     * propertyChangeEvent when a new Task of its specific type is added and the
//     * TaskService adds the Task to the TaskList and opens the overview for the
//     * newly added Task.
//     */
//    public void loadTaskList() {
//        Collection<TaskManager> taskManagerCollection = (Collection<TaskManager>) Lookup.getDefault().lookupAll(TaskManager.class);
//        for (TaskManager taskManager : taskManagerCollection) {
//            taskManager.addPropertyChangeListener(this, TaskManager.PROP_TASK_ADDED);
//            List<Task> localTaskList = taskManager.loadTasks();
//            for (Task task : localTaskList) {
//                addTask(task);
//                activateTask(task);
//            }
//        }
//    }

    public void addPropertyChangeListener(PropertyChangeListener listener) {
        propertyChange.addPropertyChangeListener(listener);
    }

    public void removePropertyCHangeListener(PropertyChangeListener listener) {
        propertyChange.removePropertyChangeListener(listener);
    }

    public void addToActiveTask(Entity entity) {
        activeTask.getContext().addItem(entity);
    }

    public void removeFromActiveTask(Entity entity) {
        activeTask.getContext().removeItem(entity);
    }

    @Override
    public void propertyChange(PropertyChangeEvent evt) {
        if (evt.getNewValue() != null) {
            addAndActivateTask((Task) evt.getNewValue());
//            ((Task) evt.getNewValue()).openTaskOverview();
        }
    }
}
